"""
Argument matching.
"""
from operator import contains, eq, ge, gt, le, lt, ne


class InvalidArg:
    def __bool__(self):
        return False


def arg_comparitor(name):
    """
    :param arg name
    :return: pair containing  name, comparitor

    given an argument name, munge it and return a proper comparitor

    >>> arg_comparitor("a")
    a, operator.eq

    >>> arg_comparitor("a__in")
    a, operator.contains
    """
    if name.endswith("__in"):
        return name[:-4], contains
    elif name.endswith("__ge"):
        return name[:-4], ge
    elif name.endswith("__gt"):
        return name[:-4], gt
    elif name.endswith("__le"):
        return name[:-4], le
    elif name.endswith("__lt"):
        return name[:-4], lt
    if name.endswith("__eq"):
        return name[:-4], eq
    if name.endswith("__ne"):
        return name[:-4], ne
    else:
        return name, eq


def arg_match(m_arg, arg, comparator=eq, default=False):
    """
    :param m_arg: value to match against or callable
    :param arg: arg to match
    :param comparator:  function that returns True if m_arg and arg match
    :param default: will be returned if m_arg is None

    if m_arg is a callable it will be called with arg

    >>> arg_match(1, 1)
    True

    >>> arg_match(1, 2)
    True

    You can match by sub args by passing in a dict

    >>> from collections import namedtuple
    >>> Msg = namedtuple("msg", ["note", "type"])
    >>> m = Msg(note=1, type="note_on")
    >>> arg_match(dict(note=1), m)
    True
    """
    if m_arg is None:
        return default
    if isinstance(m_arg, dict):
        for name, value in m_arg.items():
            name, _comparator = arg_comparitor(name)
            subarg = getattr(arg, name, InvalidArg)
            if subarg is InvalidArg:
                return subarg
            matched = arg_match(subarg, value, _comparator, default)
            if not matched or matched is InvalidArg:
                return False
        return True
    else:
        if hasattr(m_arg, "__call__"):
            return m_arg(arg)
        else:
            return comparator(arg, m_arg)


def args_match(m_args, m_kwargs, default, *args, **kwargs):
    """
    :param m_args:   values to match args against
    :param m_kwargs: values to match kwargs against
    :param default: for None, usually you would pass False
    :param arg: args to match
    :param arg: kwargs to match
    """
    if len(m_args) > len(args):
        return False
    for m_arg, arg in zip(m_args, args):
        matches = arg_match(m_arg, arg, eq)
        if not matches or matches is InvalidArg:
            return False  # bail out

    if m_kwargs:
        for name, m_arg in m_kwargs.items():
            name, comparator = arg_comparitor(name)
            arg = kwargs.get(name)
            if not arg_match(m_arg, arg, comparator, default):
                return False  # bail out
    return True



def filter_by_attr(seq, **kwargs):
    for item in seq:
        if arg_match(kwargs, item):
            yield item